@@ -13,7 +13,8 @@ import sys
 import numpy
 import threading
 
-from numexpr import interpreter, expressions, use_vml, is_cpu_amd_intel
+is_cpu_amd_intel = False # DEPRECATION WARNING: WILL BE REMOVED IN FUTURE RELEASE
+from numexpr import interpreter, expressions, use_vml
 from numexpr.utils import CacheDict
 
 # Declare a double type that does not exist in Python space
@@ -35,6 +36,38 @@ type_to_kind = expressions.type_to_kind
 kind_to_type = expressions.kind_to_type
 default_type = kind_to_type[expressions.default_kind]
 
+# VML functions that are implemented in numexpr
+vml_functions = [
+    "div",  # interp_body.cpp
+    "inv",  # interp_body.cpp
+    "pow",  # interp_body.cpp
+    # Keep the rest of this list in sync with the ones listed in functions.hpp
+    "sqrt",
+    "sin",
+    "cos",
+    "tan",
+    "arcsin",
+    "arccos",
+    "arctan",
+    "sinh",
+    "cosh",
+    "tanh",
+    "arcsinh",
+    "arccosh",
+    "arctanh",
+    "log",
+    "log1p",
+    "log10",
+    "exp",
+    "expm1",
+    "absolute",
+    "conjugate",
+    "arctan2",
+    "fmod",
+    "ceil",
+    "floor"
+    ]
+
 # Final addtions for Python 3 (mainly for PyTables needs)
 if sys.version_info[0] > 2:
     typecode_to_kind['s'] = 'str'
@@ -78,6 +111,18 @@ class ASTNode(object):
             if getattr(self, name) != getattr(other, name):
                 return False
         return True
+    
+    def __lt__(self,other):
+        # RAM: this is a fix for issue #88 whereby sorting on constants 
+        # that may be of astKind == 'complex' but type(self.value) == int or float
+        # Here we let NumPy sort as it will cast data properly for comparison 
+        # when the Python built-ins will raise an error.
+        if self.astType == 'constant':
+            if self.astKind == other.astKind:
+                return numpy.array(self.value) < numpy.array(other.value)
+            return self.astKind < other.astKind
+        else:
+            raise TypeError( 'Sorting not implemented for astType: %s'%self.astType )
 
     def __hash__(self):
         if self.astType == 'alias':
@@ -292,16 +337,20 @@ def convertConstantToKind(x, kind):
     # Exception for 'float' types that will return the NumPy float32 type
     if kind == 'float':
         return numpy.float32(x)
+    elif sys.version_info[0] >= 3 and isinstance(x,str):
+        return x.encode('ascii')
     return kind_to_type[kind](x)
 
 
 def getConstants(ast):
-    const_map = {}
-    for a in ast.allOf('constant'):
-        const_map[(a.astKind, a.value)] = a
-    ordered_constants = list(const_map.keys())
-    ordered_constants.sort()
-    constants_order = [const_map[v] for v in ordered_constants]
+    '''
+    RAM: implemented magic method __lt__ for ASTNode to fix issues
+    #88 and #209. The following test code works now, as does the test suite.
+    import numexpr as ne
+    a = 1 + 3j; b = 5.0
+    ne.evaluate( 'a*2 + 15j - b' )
+    '''
+    constants_order = sorted( ast.allOf('constant') )
     constants = [convertConstantToKind(a.value, a.astKind)
                  for a in constants_order]
     return constants_order, constants
@@ -665,15 +714,7 @@ def getExprNames(text, context):
         ex_uses_vml = False
     else:
         for node in ast.postorderWalk():
-            if node.astType == 'op' \
-                    and node.value in ['sin', 'cos', 'exp', 'log',
-                                       'expm1', 'log1p',
-                                       'pow', 'div',
-                                       'sqrt', 'inv',
-                                       'sinh', 'cosh', 'tanh',
-                                       'arcsin', 'arccos', 'arctan',
-                                       'arccosh', 'arcsinh', 'arctanh',
-                                       'arctan2', 'abs']:
+            if node.astType == 'op' and node.value in vml_functions:
                 ex_uses_vml = True
                 break
         else:
@@ -682,9 +723,28 @@ def getExprNames(text, context):
     return [a.value for a in input_order], ex_uses_vml
 
 
+def getArguments(names, local_dict=None, global_dict=None):
+    """Get the arguments based on the names."""
+    call_frame = sys._getframe(2)
+    if local_dict is None:
+        local_dict = call_frame.f_locals
+    if global_dict is None:
+        global_dict = call_frame.f_globals
+
+    arguments = []
+    for name in names:
+        try:
+            a = local_dict[name]
+        except KeyError:
+            a = global_dict[name]
+        arguments.append(numpy.asarray(a))
+    return arguments
+
+
 # Dictionaries for caching variable names and compiled expressions
 _names_cache = CacheDict(256)
 _numexpr_cache = CacheDict(256)
+_numexpr_last = {}
 
 evaluate_lock = threading.Lock()
 
@@ -732,40 +792,55 @@ def evaluate(ex, local_dict=None, global_dict=None,
             like float64 to float32, are allowed.
           * 'unsafe' means any data conversions may be done.
     """
+    global _numexpr_last
+    if not isinstance(ex, str):
+        raise ValueError("must specify expression as a string")
+    # Get the names for this expression
+    context = getContext(kwargs, frame_depth=1)
+    expr_key = (ex, tuple(sorted(context.items())))
+    if expr_key not in _names_cache:
+        _names_cache[expr_key] = getExprNames(ex, context)
+    names, ex_uses_vml = _names_cache[expr_key]
+    arguments = getArguments(names, local_dict, global_dict)
+
+    # Create a signature
+    signature = [(name, getType(arg)) for (name, arg) in
+                 zip(names, arguments)]
+
+    # Look up numexpr if possible.
+    numexpr_key = expr_key + (tuple(signature),)
+    try:
+        compiled_ex = _numexpr_cache[numexpr_key]
+    except KeyError:
+        compiled_ex = _numexpr_cache[numexpr_key] = \
+                      NumExpr(ex, signature, **context)
+    kwargs = {'out': out, 'order': order, 'casting': casting,
+              'ex_uses_vml': ex_uses_vml}
+    _numexpr_last = dict(ex=compiled_ex, argnames=names, kwargs=kwargs)
     with evaluate_lock:
-        if not isinstance(ex, str):
-            raise ValueError("must specify expression as a string")
-        # Get the names for this expression
-        context = getContext(kwargs, frame_depth=1)
-        expr_key = (ex, tuple(sorted(context.items())))
-        if expr_key not in _names_cache:
-            _names_cache[expr_key] = getExprNames(ex, context)
-        names, ex_uses_vml = _names_cache[expr_key]
-        # Get the arguments based on the names.
-        call_frame = sys._getframe(1)
-        if local_dict is None:
-            local_dict = call_frame.f_locals
-        if global_dict is None:
-            global_dict = call_frame.f_globals
-    
-        arguments = []
-        for name in names:
-            try:
-                a = local_dict[name]
-            except KeyError:
-                a = global_dict[name]
-            arguments.append(numpy.asarray(a))
-    
-        # Create a signature
-        signature = [(name, getType(arg)) for (name, arg) in zip(names, arguments)]
-    
-        # Look up numexpr if possible.
-        numexpr_key = expr_key + (tuple(signature),)
-        try:
-            compiled_ex = _numexpr_cache[numexpr_key]
-        except KeyError:
-            compiled_ex = _numexpr_cache[numexpr_key] = \
-                NumExpr(ex, signature, **context)
-        kwargs = {'out': out, 'order': order, 'casting': casting,
-                  'ex_uses_vml': ex_uses_vml}
         return compiled_ex(*arguments, **kwargs)
+
+
+def re_evaluate(local_dict=None):
+    """Re-evaluate the previous executed array expression without any check.
+
+    This is meant for accelerating loops that are re-evaluating the same
+    expression repeatedly without changing anything else than the operands.
+    If unsure, use evaluate() which is safer.
+
+    Parameters
+    ----------
+
+    local_dict : dictionary, optional
+        A dictionary that replaces the local operands in current frame.
+
+    """
+    try:
+        compiled_ex = _numexpr_last['ex']
+    except KeyError:
+        raise RuntimeError("not a previous evaluate() execution found")
+    argnames = _numexpr_last['argnames']
+    args = getArguments(argnames, local_dict)
+    kwargs = _numexpr_last['kwargs']
+    with evaluate_lock:
+        return compiled_ex(*args, **kwargs)
